/* SCALP instruction issuer */

use(channels).
use(msg).

chunk : type = record ( fu :: int,
                        instr :: int ).

fifo(stg:val(int),             /* Number of stages */
     inp:var(chan(T:type)),    /* Input */
     outp:var(chan(T)),        /* Output */
     name:val(string)          /* Name (for bview) */
     ) : expr(void) = (
      
  occ:var(int).
      
  /* Generic fifo */
  
  fifostage(inp:var(chan(T:type)),
            outp:var(chan(T))) : expr(void) =(

    /* One stage of a fifo */

    x:var(T).
    forever(
      inp?(x:=?inp);
      wait_for(5);
      outp!x
    )
  ).

  I:var(chan(T)^eval(stg-1)).  /* Inter-stage channels */
  n:var(int).
  
  newstringvalparam(name);
  forseq n in 0 to (stg-2) do (
    init_chan(I[n])
  );
  occ:=0;

  (
    ("[1]" is (
        x:var(T).
        forever(
          inp?(x:=?inp);
          occ:=occ+1;
          bview_set_text(name,i2s(occ));
          wait_for(5);
          I[0]!x
        )
      )
    ) |
    ( forpar (stg-2) do (
        ("["@i2s(forparnum+2)@"]") is fifostage(I[forparnum],I[forparnum+1])
      )
    ) |
    (("["@i2s(stg-1)@"]") is (
        x:var(T).
        forever(
          I[stg-2]?(x:=?(I[stg-2]));
          wait_for(5);
          outp!x;
          occ:=occ-1;
          bview_set_text(name,i2s(occ))
        )
      )
    )
  );

  endstringvalparam(name)
).
  

par_issuer(input:var(chan(chunk^5)),
           output:var(chan(int)^5)) : expr(void) = (
            
  /* Instruction Issuer for the SCALP Processor */
            
  chunkplus:type=record(fu::int,
                        instr::int,
                        gen::int).
  fifo_inputs:var(chan(chunkplus)^5).   /* Output of splitter */
  fifo_outputs:var(chan(chunkplus)^5).  /* Output of fifo */
  lg_outputs:var(chan(chunkplus)^5).    /* Output of last_gen updater */
  last_gen:var(int^5).      /* Generation number of instruction
                               currently visible to the controllers
                               in each column, or if none present
                               generation number of last one on that
                               channel. */

  splitter(input:var(chan(chunk^5)),
           output:var(chan(chunkplus)^5)) : expr(void) = (
  
    /* Split each FIG into five separate streams.
       Annotate each chunk with a "generation number". */
    
    gen:var(int).

    gen:=0;
    forever (
      input ? (                      /* Read FIG from input */
        forpar 5 do (                /* Send to each fifo in parallel */
          c:var(chunk).
          cp:var(chunkplus).
          c:=(?input)[forparnum];    /* Get chunk from input */
          cp->fu:=c->fu;
          cp->instr:=c->instr;
          cp->gen:=gen;                /* Add generation number */
          output[forparnum]!cp       /* Send to output */
        )
      );
      gen:=gen+1
    )
  ).

  last_gen_updater(last_gen_var:var(int),    /* Element of last_gen */
                   inp:var(chan(chunkplus)), /* Input channel */
                   outp:var(chan(chunkplus)) /* Output channel */
                   ) : expr(void) = (

    /* Update one element of the last_gen variable as channel data 
       passes through*/

    g:var(int).

    last_gen_var:=-1;
    forever (
      inp?(
        g:=(?inp)->gen;
        last_gen_var:=g;
        outp!(?inp);
        last_gen_var:=g+1
      )
    )
  ).

  controller(n:val(int),              /* Functional unit to issue to */
             inputs:var(chan(chunkplus)^5),  /* Input channels */
             last_gen:var(int^5),     /* Last gen. no. on each input */
             out:var(chan(int))       /* Output to functional unit */
             ) : expr(void) = (
    
    /* Control one functional unit */
    
    i:var(int).     /* Chunk no. being checked */
    gen:var(int).   /* Generation number currently studied */
    
    gen:=0;
    forever(
      forseq i in 0 to 4 do (
        bview_set_colour("cell"@i2s(n)@i2s(i),"yellow2");
        msg("waiting for gen "@i2s(gen)@" on column "@i2s(i));
        wait_until(last_gen[i]>gen ||
                   (chan_ready(inputs[i]) && (last_gen[i]=gen)));
        if (last_gen[i]=gen
               && chan_peek(inputs[i])->fu=n) then (
          msg("go!!");
          bview_set_colour("cell"@i2s(n)@i2s(i),"red");
          inputs[i] ? (out ! (?(inputs[i])->instr));
          wait_for(5)
        );
        wait_for(1);
        bview_set_colour("cell"@i2s(n)@i2s(i),"white")
      );
      gen:=gen+1
    )
  ).


  ( /* Body of issuer */
  
    /* Initialisation */
    (
      i:var(int).
      forseq i in 0 to 4 do (
        init_chan(fifo_inputs[i]);
        init_chan(lg_outputs[i]);
        last_gen[i]:=0
      )
    );
    init_chan(fifo_outputs[0],"Forked_ip0");
    init_chan(fifo_outputs[1],"Forked_ip1");
    init_chan(fifo_outputs[2],"Forked_ip2");
    init_chan(fifo_outputs[3],"Forked_ip3");
    init_chan(fifo_outputs[4],"Forked_ip4");
    
    (
      ("splitter" is
        splitter(input,fifo_inputs))
      
      |
      
      ("fifos" is
        forpar 5 do 
          fifo(3,fifo_inputs[forparnum],fifo_outputs[forparnum],
               "if"@i2s(forparnum)))

      |
      
      ("last_gen_updater" is
        forpar 5 do
          last_gen_updater(last_gen[forparnum],fifo_outputs[forparnum],
                           lg_outputs[forparnum]))
      
      |
      
      ("controllers" is
        forpar 5 do
          i2s(forparnum) is
            controller(forparnum,lg_outputs,last_gen,output[forparnum]))
    )
  )
).


func_unit(n:val(int),
          ch:var(chan(int))) : expr(void) = (

  /* Consume instructions from the issuer output */

  ("func unit "@i2s(n)) is
    forever (
      ch ? (
        bview_set_colour("fu"@i2s(n),"red");
        wait_for(50);
        bview_set_colour("fu"@i2s(n),"white")
      )
    )
).


source(input:var(chan(chunk^5))) : expr(void) = (

  /* Source of groups of instructions */

  i:var(chunk^5).   /* Group of instructions */
  n:var(int).       /* Sequential numbers given to instructions */
  x:var(int).
  f:var(int).       /* Function unit for instruction */
    
  n:=0;
  forever (
    forseq x in 0 to 4 do (    /* Put one instruction in each chunk */
      f:=rand(5);           /* Choose random functional unit */
      i[x]->fu:=f;
      i[x]->instr:=n;
      n:=n+1
    );
    input!i;                  /* Send group to issuer's input */
    wait_for(50)              /* Delay then repeat */
  )
).


input:var(chan(chunk^5)).     /* FIG input to issuer */
output:var(chan(int)^5).      /* Array of channels output from issuer */
bufd_output:var(chan(int)^5). /* Buffered output */


( /* Main Program */

  /* Initialisation */
  init_chan(input,"Input");
  init_chan(output[0],"Output0");
  init_chan(output[1],"Output1");
  init_chan(output[2],"Output2");
  init_chan(output[3],"Output3");
  init_chan(output[4],"Output4");
  init_chan(bufd_output[0],"To FU 0");
  init_chan(bufd_output[1],"To FU 1");
  init_chan(bufd_output[2],"To FU 2");
  init_chan(bufd_output[3],"To FU 3");
  init_chan(bufd_output[4],"To FU 4");

  
  (  
  
    ("source" is
      source(input))
      
    |
  
    ("issuer" is 
      par_issuer(input,output))
  
    |
  
    ("output_fifo" is
      forpar 5 do
        fifo(3,output[forparnum],bufd_output[forparnum],"of"@i2s(forparnum)))
      
    |
  
    ( forpar 5 do
        func_unit(forparnum,bufd_output[forparnum]))
      
  )
)  

    
